import logging
import re
from typing import Optional

from pdfminer.high_level import extract_pages
from pdfminer.layout import LAParams

import pdpc_decisions.classes
import pdpc_decisions.corpus_text.common as common
from pdpc_decisions.classes import Options, PDPCDecisionItem
from pdpc_decisions.corpus_text.common import BaseCorpusDocumentFactory

logger = logging.getLogger(__name__)


class SummaryDecisionFactory(BaseCorpusDocumentFactory):
    def __init__(self, **kwargs):
        super().__init__(laparams=LAParams(line_margin=1, char_margin=1.5), **kwargs)
        self._custom_text_containers = []
        self._paragraph_mark_containers = []
        self._paragraph_strings = []
        self._assigned_marks = []
        self._paragraph_mark = ''

    def pre_process(self):
        BaseCorpusDocumentFactory.pre_process(self)
        for page in range(len(self._pages)):
            page_containers = list(self.get_text_containers([page]))
            page_containers = sorted(page_containers, key=lambda item: item.x0)
            page_containers = sorted(page_containers, key=lambda container: container.y0, reverse=True)
            self._custom_text_containers.append(page_containers)
        self._custom_text_containers[0] = self._custom_text_containers[0][1:]

    def process_all(self):
        for index, page_containers in enumerate(self._custom_text_containers):
            self.process_page(page_containers)

    def process_paragraph(self, paragraph, index, page_containers):
        container_string = paragraph.get_text().strip()
        logger.info(f"New container: {container_string}")
        if match := re.match(r'^\d+\.\s*', container_string):
            self._paragraph_mark = match.group(0).strip()
            container_string = container_string.replace(match.group(0), '', 1).strip()
            logger.info(
                f'Matched a paragraph mark: {self._paragraph_mark} and adjusted container string: {container_string}')
        self._paragraph_strings.append(container_string)
        logger.info(f"Add string to paragraph.")
        if re.search(r'[.?!]$', container_string) and (
                len(self._paragraph_strings) == 1 or common.check_gap_before_after_container(page_containers, index)):
            if self._paragraph_mark:
                self._result.add_paragraph(" ".join(self._paragraph_strings).strip(), self._paragraph_mark)
                logger.info(
                    f"Added a new paragraph: ({self._paragraph_mark}) {' '.join(self._paragraph_strings).strip()}")
            else:
                logger.info('No paragraph mark found.')
                if len(self._result.paragraphs) > 0:
                    logger.info('Appending to previous paragraph.')
                    self._paragraph_strings.insert(0, self._result.paragraphs[-1].text)
                    self._result.paragraphs[-1].update_text((" ".join(self._paragraph_strings)))
                else:
                    logger.info('Creating new, unmarked paragraph.')
                    self._result.add_paragraph(" ".join(self._paragraph_strings))
            self._reset()
        logger.info('End of this container')

    def _reset(self):
        self._paragraph_strings.clear()
        self._paragraph_mark = ''
        logger.info('Reset strings and mark')

    @classmethod
    def check_decision(cls, item: Optional[PDPCDecisionItem] = None, options: Optional[Options] = None) -> bool:
        with pdpc_decisions.classes.PDFFile(item, options) as pdf:
            first_page = extract_pages(pdf, page_numbers=[0])
            containers = common.extract_text_containers(first_page)
            for container in containers:
                if container.get_text().strip() == 'SUMMARY OF THE DECISION':
                    return True
        return False



import io
import logging
import re
from collections import Counter
from typing import List, Generator, Dict, Optional

from pdfminer import layout
from pdfminer.high_level import extract_text_to_fp, extract_pages
from pdfminer.layout import LTTextContainer, LTChar, LTPage, LTTextLineHorizontal, LTRect

from pdpc_decisions.classes import Options, PDPCDecisionItem, CorpusDocument, PDFFile

logger = logging.getLogger(__name__)


def check_text_is_date(paragraph: LTTextContainer) -> bool:
    """Returns true if the text of a paragraph contains only a date."""
    return re.match(r'\d\d? \w+ \d\d\d\d\s*$', paragraph.get_text().strip()) is not None


def get_text_margins(text_containers: List[LTTextContainer], limit: int = 1) -> List[int]:
    """
    Returns a list of integers (from left to right) representing the most commonly used text
    margins in a document.

    :param limit: If there are any margins with counts less than limit, it is removed.
    :param text_containers: The document to evaluate.
    :return: A list of rounded integers.
    """
    containers_x0 = [round(container.x0) for container in text_containers]
    c = Counter(containers_x0)
    logger.info(f"Text margins: {c}")
    return sorted([margin for margin, count in c.most_common() if count > limit])


def get_main_text_size(text_containers: List[LTTextContainer]) -> int:
    """Returns the text size in points of the most common text in text containers."""
    font_sizes = [container.height for container in text_containers]
    c = Counter(font_sizes)
    size, _ = c.most_common(1)[0]
    return round(size)


def check_first_page_is_cover(pdf: bytes) -> bool:
    """Reads pdf and returns True if it is a cover page"""
    with io.StringIO() as test_string:
        params = layout.LAParams(line_margin=2)
        extract_text_to_fp(pdf, test_string, page_numbers=[0], laparams=params)
        first_page = test_string.getvalue()
        return len(first_page.split()) <= 100


def get_common_font_from_pages(pages: List[LTPage]) -> str:
    """Gets the most common font from a list of pages and return its name"""
    font_names = []
    text_containers = extract_text_containers((page for page in pages))
    for container in text_containers:
        for line in container:
            for char in line:
                if isinstance(char, layout.LTChar):
                    font_names.append(char.fontname)
    c = Counter(font_names)
    font, _ = c.most_common(1)[0]
    return font


def check_common_font(paragraph: layout.LTTextContainer, common_font: str) -> bool:
    """Returns true if the most common font in paragraph is common font, or if common
    font is one of the fonts used in the paragraph. """
    fonts = []
    if isinstance(paragraph, LTTextLineHorizontal):
        fonts.extend([char.fontname for char in paragraph if isinstance(char, LTChar)])
    else:
        for line in paragraph:
            fonts.extend([char.fontname for char in line if isinstance(char, LTChar)])
    c = Counter(fonts)
    font, _ = c.most_common(1)[0]
    return common_font == font or common_font in c


def check_gap_before_after_container(containers: List[LTTextContainer], index: int, equal=False) -> bool:
    """Returns true if the gap before the container is smaller than the gap after.
    Tries to check for the next container if there are two on the same line, or if it is the last
    container of a page."""
    index_before = index - 1
    while not index_before < 0 and containers[index].y0 == containers[index_before].y0:
        index_before -= 1
    if index_before < 0:
        return False
    index_after = index + 1
    while not index_after >= len(containers) and containers[index].y0 == containers[index_after].y0:
        index_after += 1
    if index_after == len(containers):
        return True
    gap_before = round(containers[index_before].y1 - containers[index].y0)
    gap_after = round(containers[index].y1 - containers[index_after].y0)
    if equal:
        return gap_after >= gap_before
    else:
        return gap_after > gap_before


def extract_text_containers(pages: Generator[layout.LTPage, None, None]) -> List[layout.LTTextContainer]:
    result = []
    for page in pages:
        result.extend([element for element in page if
                       isinstance(element, layout.LTTextContainer) and element.get_text() != '' and not re.search(
                           r'^\s+$', element.get_text())])
    return result


def split_joined_text_containers(containers: List[LTTextContainer]) -> List[LTTextContainer]:
    """Searches for text containers in individual containers, splits them and returns the result."""
    result = containers.copy()
    container_length = len(containers) - 1
    logger.info(f'Start splitting containers in reverse fashion.')
    for index in range(container_length, -1, -1):
        container = containers[index]
        if len(container) > 1:
            new_item = []
            for sub_container in container:
                if isinstance(sub_container, LTTextContainer) and sub_container.get_text().strip() != '':
                    logger.info(f'Found split container: {sub_container}')
                    new_item.append(sub_container)
            if len(new_item) != 0:
                result.remove(container)
                result[index:index] = new_item
                logger.info(f'Splitting joined container.')
    return result


def get_footnotes_using_separator(page: LTPage, text_containers: List[LTTextContainer],
                                  left_margin: int, main_text_size: int) \
        -> (List[LTTextContainer], Optional[List[LTTextContainer]]):
    if footnote_line_container := [container for container in page if all([
        isinstance(container, LTRect),
        container.height < 1,
        container.y0 < 350,
        container.width < 150,
        round(container.x0) == left_margin
    ])]:
        footnote_line_container = sorted(footnote_line_container, key=lambda container: container.y0)
        footnote_line = footnote_line_container[0].y0
        footnotes_page = [container for container in text_containers if all([
            container.y0 < footnote_line,
            round(container.x0) < 284,
            get_main_text_size([container]) < main_text_size])]
        text_containers = [container for container in text_containers if container not in footnotes_page]
        return text_containers, footnotes_page
    else:
        return text_containers, None


def get_endnotes_using_separator(page: LTPage, text_containers: List[LTTextContainer],
                                 left_margin: int, main_text_size: int) \
        -> (List[LTTextContainer], Optional[List[LTTextContainer]]):
    if endnote_line_container := [container for container in page if all([
        isinstance(container, LTRect),
        container.height < 1,
        container.width < 150,
        round(container.x0) == left_margin
    ])]:
        endnote_line_container = sorted(endnote_line_container, key=lambda container: container.y0)
        endnote_line = endnote_line_container[0].y0
        endnotes = [container for container in text_containers if all([
            container.y0 < endnote_line,
            get_main_text_size([container]) < main_text_size])]
        text_containers = [container for container in text_containers if container not in endnotes]
        return text_containers, endnotes
    else:
        return text_containers, None


def construct_footnotes(footnotes: List[LTTextContainer]) -> Dict[str, str]:
    footnote_text = []
    result = []
    for container in footnotes:
        footnote_string = container.get_text().strip()
        footnote_text.append(footnote_string)
        if re.search(r'[.?!"”]\s*$', footnote_string):
            if footnote_mark := re.match(r'\d+\s+', footnote_text[0].strip()):
                footnote_mark = footnote_mark.group(0).strip()
                footnote = ' '.join(footnote_text).replace(footnote_mark, '', 1).strip()
                result.append((footnote_mark, footnote))
                logger.info(f"Found a footnote: ({footnote_mark}, {footnote})")
            else:
                logger.warning(f'No footnote mark found: {footnote_text[0].strip()}')
                if len(result) > 0:
                    logger.info(f'Appending to previous footnote.')
                    footnote_text.insert(0, result[-1][1])
                    footnote = ' '.join(footnote_text).strip()
                    result[-1] = (result[-1][0], footnote)
                else:
                    logger.error(f"Nowhere to put footnote. Abandoning: {' '.join(footnote_text)}")
            footnote_text.clear()
    return dict(result)


class BaseCorpusDocumentFactory:
    """Base class for a factory to produce CorpusDocuments. Provides a basic extraction of text containers from PDF."""

    def __init__(self, **kwargs):
        self.data: Dict[str, any] = dict(kwargs)
        self._result = CorpusDocument(source=self.data.get('source', None))
        self._text_containers: List[List[LTTextContainer]] = []
        self._pages: List[LTPage] = []

    def pre_process(self):
        """Actions to do be fore each paragraph is processed. Use this to initialize metadata
        or clean text containers."""
        if pdf := self.data.get('pdf', None):
            with open(pdf, 'rb') as pdf:
                self._extract_pages_and_text_containers(pdf)
        elif source := self.data.get('source', None):
            with PDFFile(source, self.data.get('options', None)) as file:
                self._extract_pages_and_text_containers(file)

    def process_all(self):
        """Process all pages. This is the entry point to processing pages and paragraphs."""
        for index, page_containers in enumerate(self._text_containers):
            self.process_page(page_containers)

    def process_page(self, page_containers: List[LTTextContainer]):
        """Process a page. This is the entry point to processing a page and its paragraphs."""
        for index, container in enumerate(page_containers):
            self.process_paragraph(container, index, page_containers)

    def process_paragraph(self, paragraph: LTTextContainer, index: int, page_containers: List[LTTextContainer]):
        """Process a paragraph. The default add a paragraph to the CorpusDocument"""
        self._result.add_paragraph(paragraph.get_text().strip(), str(index))

    def get_result(self) -> CorpusDocument:
        """Convenience method to call all processing methods and return the result CorpusDocument."""
        self.pre_process()
        self.process_all()
        return self._result

    def _extract_pages_and_text_containers(self, pdf):
        """Method called to prepare the pages and containers from pdf for processing during pre-process."""
        self._pages = list(extract_pages(pdf, laparams=self.data.get('laparams', None)))
        for page in self._pages:
            self._text_containers.append([element for element in page if
                                          isinstance(element, LTTextContainer) and
                                          element.get_text() != '' and not
                                          re.search(r'^\s+$', element.get_text())])

    def get_text_containers(self, page_numbers: Optional[List[int]] = None, filter_function=lambda container: True,
                            custom_container: List[List[LTTextContainer]] = None) \
            -> Generator[LTTextContainer, None, None]:
        """
        Convenience method to generate a list of containers from the source/PDF.

        :param page_numbers: Array of zero-indexed pages to extract. Default is to extract from all pages.
        :param filter_function: Function to call to filter extracted text containers.
            Function accepts one parameter being the container to retain or reject.
            Function should return a boolean (True if the container should be retained)
        :param custom_container: The containers which should be processed.
            Default is the main text containers in the factory.
        :return:
        """
        if custom_container is None:
            custom_container = self._text_containers
        if not page_numbers:
            for page in custom_container:
                for container in page:
                    if filter_function(container):
                        yield container
        else:
            for index in page_numbers:
                for page in custom_container[index]:
                    for container in page:
                        if filter_function(container):
                            yield container

    @classmethod
    def from_pdf(cls, pdf, **kwargs) -> CorpusDocument:
        """Class method to create a CorpusDocument from a PDF."""
        factory = cls(pdf=pdf, **kwargs)
        return factory.get_result()

    @classmethod
    def from_decision_item(cls, source: PDPCDecisionItem, options: Options = None, **kwargs) -> CorpusDocument:
        """Class method to create a CorpusDocument from a PDPCDecisionItem."""
        factory = cls(source=source, options=options, **kwargs)
        return factory.get_result()

    @classmethod
    def check_decision(cls, item: Optional[PDPCDecisionItem] = None, options: Optional[Options] = None) -> bool:
        """Class method to check if the Item should be handled by this Factory.
        Returns false for BaseCorpusDocumentFactory."""
        return False



import logging
import re
from typing import List, Dict, Optional

from pdfminer.high_level import extract_pages, extract_text
from pdfminer.layout import LTTextContainer, LAParams, LTTextBoxHorizontal, LTTextLineHorizontal, LTChar

from pdpc_decisions.classes import PDPCDecisionItem, Options, PDFFile
from pdpc_decisions.corpus_text import common
from pdpc_decisions.corpus_text.common import BaseCorpusDocumentFactory

logger = logging.getLogger(__name__)


def check_next_footnote(footnotes, index):
    if not len(footnotes) == index + 1:
        next_footnote = footnotes[index + 1]
    else:
        return False
    if re.match(r'\d+\s*', next_footnote.get_text().strip()):
        if isinstance(next_footnote, LTTextBoxHorizontal):
            first_char = next(iter(next(iter(next_footnote))))
        elif isinstance(next_footnote, LTTextLineHorizontal):
            first_char = next(iter(next_footnote))
        else:
            return False
        if round(first_char.height) < common.get_main_text_size([next_footnote]):
            return True
        else:
            return False


class DecisionV2Factory(BaseCorpusDocumentFactory):
    def __init__(self, **kwargs):
        super(DecisionV2Factory, self).__init__(laparams=LAParams(line_margin=0.5), **kwargs)
        self._common_font = ''
        self._paragraph_marks = []
        self._text_margins = [72]
        self._paragraph_strings: List[str] = []
        self._current_paragraph_mark = ''
        self._main_text_size = 12
        self._footnotes: Dict[str, str] = dict()

    def _extract_pages_and_text_containers(self, pdf):
        self._pages = list(extract_pages(pdf, laparams=self.data.get('laparams', None)))[1:]
        for page in self._pages:
            self._text_containers.append([element for element in page if
                                          isinstance(element, LTTextContainer) and
                                          element.get_text() != '' and not
                                          re.search(r'^\s+$', element.get_text())])

    def pre_process(self):
        BaseCorpusDocumentFactory.pre_process(self)
        self._common_font = common.get_common_font_from_pages(self._pages)
        margin_limit = 3 if len(self._pages) > 3 else len(self._pages)
        self._text_margins = common.get_text_margins(list(self.get_text_containers()), margin_limit)
        self._main_text_size = common.get_main_text_size(list(self.get_text_containers()))
        for index, containers in enumerate(self._text_containers):
            containers = common.split_joined_text_containers(containers)
            containers = [container for container in containers if all([
                container.get_text().strip() != '',
                round(container.x0) in self._text_margins,
                round(container.height) <= self._main_text_size + 1 or self._check_footnote(container)
            ])]
            containers = sorted(containers, key=lambda item: item.x0)
            containers = sorted(containers, key=lambda item: item.y0, reverse=True)
            self._text_containers[index] = containers
        self._footnotes = self._get_and_remove_footnotes()
        logger.info("Pre process finished.")

    def _get_and_remove_footnotes(self):
        footnotes = []
        for index, page in enumerate(self._pages):
            new_containers, footnote_page = common.get_footnotes_using_separator(page,
                                                                                 self._text_containers[index],
                                                                                 self._text_margins[0],
                                                                                 self._main_text_size)
            self._text_containers[index] = new_containers
            if footnote_page:
                footnotes.extend(footnote_page)
        return common.construct_footnotes(footnotes)

    def process_all(self):
        for index, page_containers in enumerate(self._text_containers):
            self.process_page(page_containers)

    def process_paragraph(self, paragraph: LTTextContainer, index: int, page_containers: List[LTTextContainer]) -> None:
        logger.info(f"New container: {paragraph}")
        if self._check_skip_paragraph(paragraph, index):
            return
        container_string = self._replace_footnote(paragraph)
        container_string = self._check_top_paragraph(container_string, paragraph)
        self._paragraph_strings.append(container_string)
        logger.info(f'Added to paragraph strings')
        self._check_paragraph_end(index, page_containers)
        logger.info("End of this container.")

    def _check_footnote(self, paragraph):
        result = []
        char_text = []
        if isinstance(paragraph, LTTextBoxHorizontal):
            char_list = list(iter(next(iter(paragraph))))
        elif isinstance(paragraph, LTTextLineHorizontal):
            char_list = list(iter(paragraph))
        else:
            return None
        for index, char in enumerate(char_list):
            if isinstance(char, LTChar) and round(char.height) < self._main_text_size and \
                    re.match(r'\d', char.get_text()):
                char_text.append(char.get_text())
                if (index + 1 == len(char_list)) or (isinstance(char_list[index + 1], LTChar) and round(
                        char_list[index + 1].height) >= self._main_text_size):
                    footnote_mark = ''.join(char_text)
                    result.append(footnote_mark)
                    char_text.clear()
        return result if len(result) > 0 else None

    def _replace_footnote(self, paragraph):
        result = paragraph.get_text().strip()
        if footnotes_marks := self._check_footnote(paragraph):
            for footnote_mark in footnotes_marks:
                if self._footnotes.get(footnote_mark):
                    result = result.replace(footnote_mark, f" ({self._footnotes[footnote_mark]})", 1)
                    logger.info(f'Replaced a footnote: {footnote_mark}, {self._footnotes[footnote_mark]}')
                else:
                    logger.warning(f'Footnote mark ({footnote_mark}) cannot be replaced as it is not in the footnotes.')
        return result

    def _check_top_paragraph(self, container_string, paragraph):
        if all([
            round(paragraph.x0) == self._text_margins[0],
            match := re.match(r'\d+\.?\s*', paragraph.get_text()),
            not self._current_paragraph_mark
        ]):
            self._current_paragraph_mark = match.group(0).strip()
            logger.info(f"Added a paragraph mark: {self._current_paragraph_mark}")
            container_string = container_string.replace(self._current_paragraph_mark, '', 1)
            logger.info(f"Adjust container: {container_string}")
        return container_string

    def _check_paragraph_end(self, index, page_containers):
        if all([common.check_gap_before_after_container(page_containers, index, equal=True),
                not common.check_common_font(page_containers[index], self._common_font),
                not self._current_paragraph_mark
                ]):
            self._result.add_paragraph(" ".join(self._paragraph_strings))
            logger.info(f'Added a header-like paragraph: {self._result.paragraphs[-1].text}')
            self._paragraph_strings.clear()
            return
        if re.search(r'[.?!][")]?\d*\s*$', page_containers[index].get_text()) and any([
            len(self._paragraph_strings) == 1,
            common.check_gap_before_after_container(page_containers, index)
        ]):
            if self._current_paragraph_mark:
                self._result.add_paragraph(" ".join(self._paragraph_strings), self._current_paragraph_mark)
            else:
                logger.warning(
                    f'No paragraph mark was found for ({self._paragraph_strings[0]}).')
                if len(self._result.paragraphs) > 0:
                    logger.info('Adding to previous parargraph')
                    self._paragraph_strings.insert(0, self._result.paragraphs[-1].text)
                    self._result.paragraphs[-1].update_text(" ".join(self._paragraph_strings))
                else:
                    logger.warning('Creating a new paragraph')
                    self._result.add_paragraph(" ".join(self._paragraph_strings))
            logger.info(f'Added a paragraph: {self._result.paragraphs[-1]}')
            self._paragraph_strings.clear()
            self._current_paragraph_mark = ''
            logger.info('Reset paragraph mark and string.')

    def _check_skip_paragraph(self, paragraph, index):
        paragraph_text = paragraph.get_text().strip()
        if common.check_text_is_date(paragraph):
            logger.info('Date found, skipping')
            return True
        if any([
            re.match(r'[A-Z ]+\s*$', paragraph_text),
            re.match(r'(\[\d{4}])\s+((?:\d\s+)?[A-Z|()]+)\s+\[?(\d+)\]?', paragraph_text),
            re.match(r'Tan Kiat How', paragraph_text),
            re.match(r'Yeong Zee Kin', paragraph_text)
        ]):
            logger.info('Meta-info found, skipping')
            return True
        if index == 0 and len(self._pages) > 1:
            if paragraph_text == self._text_containers[1][0].get_text().strip():
                logger.info('Looks like a header. Skipping.')
                return True
        if round(paragraph.y0) > 700 and re.match(r'(\[\d{4}])\s+((?:\d\s+)?[A-Z|()]+)\s+\[?(\d+)\]?', paragraph_text):
            logger.info('Looks like a header. Skipping.')
            return True
        if 284 < round(paragraph.x0) < 296 and re.match(r'\d+$', paragraph_text):
            logger.info('Looks like a footer. Skipping.')
            return True
        if re.match(r'\d+ of \d+$', paragraph_text):
            logger.info('Looks like a footer. Skipping.')
            return True
        return False

    @classmethod
    def check_decision(cls, item: Optional[PDPCDecisionItem] = None, options: Optional[Options] = None) -> bool:
        with PDFFile(item, options) as pdf:
            first_page = extract_pages(pdf, page_numbers=[0], laparams=LAParams(line_margin=0.1, char_margin=3.5))
            text = extract_text(pdf, page_numbers=[0], laparams=LAParams(line_margin=0.1, char_margin=3.5))
            containers = common.extract_text_containers(first_page)
            if len(text.split()) <= 100:
                for container in containers:
                    container_text = container.get_text().strip()
                    if container_text == 'DECISION' or container_text == 'GROUNDS OF DECISION':
                        return True
        return False



import logging
import re
from typing import List, Dict, Optional

from pdfminer.high_level import extract_pages, extract_text
from pdfminer.layout import LAParams, LTTextContainer, LTRect, LTChar, LTTextBoxHorizontal, LTTextLineHorizontal

import pdpc_decisions.classes
import pdpc_decisions.corpus_text.common as common
from pdpc_decisions.classes import PDPCDecisionItem, Options

logger = logging.getLogger(__name__)


class DecisionV1Factory(common.BaseCorpusDocumentFactory):
    def __init__(self, **kwargs):
        super().__init__(laparams=LAParams(line_margin=0.1, char_margin=3.5), **kwargs)
        self._custom_text_containers: List[List[LTTextContainer]] = []
        self._paragraph_marks = []
        self._text_margins = [72]
        self._paragraph_strings: List[str] = []
        self._assigned_marks = []
        self._current_paragraph_mark = ''
        self._main_text_size = 12
        self._footnotes: Dict[str, str] = dict()

    def _extract_pages_and_text_containers(self, pdf):
        self._pages = list(extract_pages(pdf, laparams=self.data.get('laparams', None)))
        for page in self._pages:
            self._text_containers.append([element for element in page if
                                          isinstance(element, LTTextContainer) and all([
                                              element.get_text().strip() != '',
                                          ])])

    def pre_process(self):
        common.BaseCorpusDocumentFactory.pre_process(self)
        self._text_margins = common.get_text_margins(
            list(self.get_text_containers()), len(self._pages))
        self._get_custom_text_containers()
        self._main_text_size = common.get_main_text_size(
            list(self.get_text_containers(custom_container=self._custom_text_containers)))
        self._footnotes = self._get_and_remove_footnotes()

    def _get_custom_text_containers(self):
        logger.info('Preparing filtered text containers')
        self._custom_text_containers.clear()
        for index, page_containers in enumerate(self._text_containers):
            page_containers = [container for container in page_containers if
                               round(container.x0) in self._text_margins]
            page_containers = common.split_joined_text_containers(page_containers)
            page_containers = self._remove_underlined_headers_from_page(page_containers, index)
            page_containers = sorted(page_containers, key=lambda item: item.x0)
            page_containers = sorted(page_containers, key=lambda container: container.y0, reverse=True)
            self._custom_text_containers.append(page_containers)

    def _remove_underlined_headers_from_page(self, page_containers, index):
        underlined_headers = [element for element in self._pages[index] if isinstance(element, LTRect)
                              and round(element.x0) == self._text_margins[0]]
        page_containers_result = page_containers.copy()
        for header in underlined_headers:
            for container in page_containers:
                if container.y0 < header.y0 <= container.y1 and container.get_text().strip() != '':
                    logger.info(f'Found an underlined header: {container.get_text()}')
                    page_containers_result.remove(container)
        return page_containers_result

    def _get_and_remove_footnotes(self) -> Dict[str, str]:
        footnotes = []
        for index, page in enumerate(self._pages):
            new_containers, footnote_page = common.get_footnotes_using_separator(page,
                                                                                 self._custom_text_containers[index],
                                                                                 self._text_margins[0],
                                                                                 self._main_text_size)
            self._custom_text_containers[index] = new_containers
            if footnote_page:
                footnotes.extend(footnote_page)
        result = common.construct_footnotes(footnotes)
        if not result:
            new_containers, endnotes = common.get_endnotes_using_separator(self._pages[-1],
                                                                           self._custom_text_containers[-1],
                                                                           self._text_margins[0],
                                                                           self._main_text_size)
            self._custom_text_containers[-1] = new_containers
            if endnotes:
                result = common.construct_footnotes(endnotes)
        return result

    def process_all(self):
        for index, page_containers in enumerate(self._custom_text_containers):
            self.process_page(page_containers)

    def process_paragraph(self, paragraph, index, page_containers):
        logger.info(f"New container: {paragraph}")
        if common.check_text_is_date(paragraph):
            logger.info('Date found, skipping')
            return
        if re.match(r'Case Number:', paragraph.get_text()):
            logger.info('Metadata found, skipping.')
            return
        container_string = self._replace_footnote(paragraph)
        container_string = self._check_top_paragraph(container_string)
        self._paragraph_strings.append(container_string)
        logger.info(f'Added to paragraph strings')
        self._check_paragraph_end(container_string, index, page_containers)
        logger.info("End of this container.")

    def _check_paragraph_end(self, container_string, index, page_containers):
        def set_paragraph():
            if self._current_paragraph_mark:
                self._result.add_paragraph(" ".join(self._paragraph_strings), self._current_paragraph_mark)
                logger.info(f'Added a paragraph: {" ".join(self._paragraph_strings)}')
            else:
                logger.warning('No paragraph mark was found.')
                if len(self._result.paragraphs) > 0:
                    logger.info('Appending to last paragraph.')
                    self._paragraph_strings.insert(0, self._result.paragraphs[-1].text)
                    self._result.paragraphs[-1].update_text(" ".join(self._paragraph_strings))
                    logger.info(f'Replaced a paragraph: {" ".join(self._paragraph_strings)}')
                else:
                    logger.info('Adding a new paragraph with no mark.')
                    self._result.add_paragraph(" ".join(self._paragraph_strings))
            self._paragraph_strings.clear()
            self._current_paragraph_mark = ''
            logger.info('Reset paragraph mark and string.')

        if re.search(r'[.?!][")]?\d*\s*$', container_string) and any([
            len(self._paragraph_strings) == 1,
            common.check_gap_before_after_container(page_containers, index, equal=True)
        ]):
            set_paragraph()
        elif container_string.isupper():
            if len(page_containers) > index + 1 and page_containers[index + 1].get_text().isupper() is False:
                set_paragraph()

    def _check_top_paragraph(self, container_string):
        if match := re.match(r'[\dA-Z]+\.\s*', container_string):
            self._current_paragraph_mark = match.group(0).strip()
            logger.info(f"Added a paragraph mark: {self._current_paragraph_mark}")
            container_string = container_string.replace(self._current_paragraph_mark, '', 1).strip()
            logger.info(f"Adjust container: {container_string}")
        return container_string

    def _replace_footnote(self, container):
        char_text = []
        result = container.get_text().strip()
        if isinstance(container, LTTextBoxHorizontal):
            char_list = list(iter(next(iter(container))))
        elif isinstance(container, LTTextLineHorizontal):
            char_list = list(iter(container))
        else:
            return result
        for index, char in enumerate(char_list):
            if isinstance(char, LTChar) and all([
                round(char.height) < self._main_text_size,
                re.match(r'\d', char.get_text())
            ]):
                char_text.append(char.get_text())
                if isinstance(char_list[index + 1], LTChar) and \
                        round(char_list[index + 1].height) >= self._main_text_size:
                    footnote_mark = ''.join(char_text)
                    if footnote_text := self._footnotes.get(footnote_mark, None):
                        result = result.replace(footnote_mark, f" ({footnote_text})", 1)
                        logger.info(f'Replaced a footnote: {footnote_mark}, {footnote_text}')
                    else:
                        logger.warning(f'No footnote was present for footnote {footnote_mark}.')
        return result

    @classmethod
    def check_decision(cls, item: Optional[PDPCDecisionItem] = None, options: Optional[Options] = None) -> bool:
        with pdpc_decisions.classes.PDFFile(item, options) as pdf:
            first_page = extract_pages(pdf, page_numbers=[0], laparams=LAParams(line_margin=0.1, char_margin=3.5))
            text = extract_text(pdf, page_numbers=[0], laparams=LAParams(line_margin=0.1, char_margin=3.5))
            containers = common.extract_text_containers(first_page)
            if len(text.split()) >= 100:
                for container in containers:
                    if container.get_text().strip() == 'GROUNDS OF DECISION':
                        return True
        return False



import logging
import re
from typing import List

from pdfminer.high_level import extract_pages
from pdfminer.layout import LAParams, LTTextContainer

import pdpc_decisions.corpus_text.common as common
from pdpc_decisions.corpus_text.common import BaseCorpusDocumentFactory

logger = logging.getLogger(__name__)


class BasicCorpusFactory(BaseCorpusDocumentFactory):
    """
    Basic factory for CorpusDocument with the aim of producing results similar to previous
    versions of pdpc-decisions, with a focus on non-destructive rendering of documents
    and ensuring that a CorpusDocument is produced with no exceptions.
    """

    def __init__(self, **kwargs):
        super().__init__(laparams=LAParams(line_margin=2), **kwargs)
        self._paragraph_strings: List[str] = []

    def _extract_pages_and_text_containers(self, pdf):
        self._pages = list(extract_pages(pdf, laparams=self.data.get('laparams', None)))
        if common.check_first_page_is_cover(pdf):
            self._pages = self._pages[1:]
        for page in self._pages:
            containers = [element for element in page if isinstance(element, LTTextContainer) and
                          element.get_text() != '' and not
                          re.search(r'^\s+$', element.get_text())]
            containers = common.split_joined_text_containers(containers)
            containers = sorted(containers, key=lambda item: item.x0)
            containers = sorted(containers, key=lambda item: item.y0, reverse=True)
            self._text_containers.append(containers)

    def process_paragraph(self, paragraph, index, page_containers):
        logger.info(f"New container: {paragraph}")
        text = paragraph.get_text().strip()
        if common.check_text_is_date(paragraph):
            logger.info('Date found, skipping')
            return
        if re.search(r'^\s+\[\d{4}]\s+(?:\d\s+)?[A-Z|()]+\s+\d+[\s.]?\s+$', text):
            logger.info('Citation found, skipping')
            return
        if re.search(r'^\d*[\s.]*$', text):
            logger.info('Number found, skipping')
            return
        self._paragraph_strings.append(text)
        logger.info(f'Added to paragraph strings')
        self._check_paragraph_end()
        logger.info("End of this container.")

    def _check_paragraph_end(self):
        text = self._paragraph_strings[-1]
        if re.search(r'\.\s*$', text):
            self._result.add_paragraph(' '.join(self._paragraph_strings))
            logger.info(f'Added a paragraph: {self._result.paragraphs[-1]}')
            self._paragraph_strings.clear()



#  MIT License Copyright (c) 2020. Houfu Ang
import logging
import os
from typing import Sequence, List, Type

from pdpc_decisions.classes import Options, PDPCDecisionItem, CorpusDocument
from pdpc_decisions.corpus_text.basic import BasicCorpusFactory
from pdpc_decisions.corpus_text.common import BaseCorpusDocumentFactory
from pdpc_decisions.corpus_text.decision_v1 import DecisionV1Factory
from pdpc_decisions.corpus_text.decision_v2 import DecisionV2Factory
from pdpc_decisions.corpus_text.summary import SummaryDecisionFactory

logger = logging.getLogger(__name__)

factory_list: List[Type[BaseCorpusDocumentFactory]] = \
    [SummaryDecisionFactory, DecisionV1Factory, DecisionV2Factory, BasicCorpusFactory]


def get_document(options: Options, item: PDPCDecisionItem) -> CorpusDocument:
    """
    Gets a CorpusDocument from a PDPCDecisionItem.
    If 'extra_corpus' option is enabled in options, it will return a CorpusDocument
    using experimental versions of CorpusDocumentFactory.

    :param options: Options passed from command line script.
    :param item: The PDPCDecisionItem to get a document from.
    :return: A CorpusDocument.
    """
    from urllib.parse import urlparse
    url = urlparse(item.download_url)
    if url.path[-3:] == 'pdf':
        if options.get('extra_corpus'):
            for factory_class in factory_list:
                if factory_class.check_decision(item, options):
                    return factory_class.from_decision_item(item, options)
        else:
            return BasicCorpusFactory.from_decision_item(item, options)


def create_corpus_file(options: Options, item: PDPCDecisionItem) -> str:
    """
    Writes a corpus file in the folder as specified in options. Checks if there is already an
    existing file in the folder and renames accordingly to prevent overwriting.

    :param options: Options passed from command line script.
    :param item: The PDPCDecisionItem to get a document from.
    :return: The path of the destination file.
    """
    logger.info(f'Now working on: {item}')
    logger.info(f'Source file: {item.download_url}')
    text = get_document(options, item).get_text_as_paragraphs()
    destination_filename = f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent}.txt"
    corpus_folder_ = options["corpus_folder"]
    destination = os.path.join(corpus_folder_, destination_filename)
    if os.path.isfile(
            os.path.join(corpus_folder_, f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent} (1).txt")):
        logger.warning('Found secondary format for corpus files. To follow format for this file.')
        file_num = 2
        destination = os.path.join(corpus_folder_,
                                   f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent} ({file_num}).txt")
        while os.path.isfile(destination):
            file_num += 1
            destination = os.path.join(corpus_folder_,
                                       f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent} ({file_num}).txt")
    if os.path.isfile(destination):
        logger.warning('Found two decisions with the same corpus output file name. Rewriting to secondary format.')
        file_num = 1
        os.rename(destination,
                  os.path.join(corpus_folder_,
                               f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent} ({file_num}).txt"))
        file_num += 1
        destination = os.path.join(corpus_folder_,
                                   f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent} ({file_num}).txt")
    with open(destination, 'w') as fOut:
        text = [line + '\n' for line in text]
        fOut.writelines(text)
    logger.info("Wrote: {}".format(destination))
    return destination


def create_corpus(options: Options, items: Sequence[PDPCDecisionItem]) -> None:
    """
    Creates a corpus from a series of PDPCDecisionItem.
    The documents are saved in a folder specified by options["corpus_folder"]
    and is created if it does not exist yet.

    :param options: Options passed from command line script.
    :param items: The PDPCDecisionItem to get a document from.
    :return: None
    """
    logger.info('Now creating corpus.')
    if not os.path.exists(options["corpus_folder"]):
        os.mkdir(options["corpus_folder"])
    for item in items:
        create_corpus_file(options, item)
    logger.info(f'Number of items in corpus: {len(items)}')
    logger.info(f'Finished creating corpus at {options["corpus_folder"]}')
